#include "appchooser.h"
#include <QUrl>
#include <QMimeDatabase>
#include <QMimeType>
#include <QStandardPaths>
#include <QDebug>
#include <gio/gdesktopappinfo.h>
#include <gio/gappinfo.h>

namespace {

void eachAppInfoCallback(gpointer data, gpointer userData)
{
    std::vector<std::string> *dstAppDesktopNameList =
        static_cast<std::vector<std::string> *>(userData);

    GAppInfo *appInfo = static_cast<GAppInfo *>(data);

    if (!G_IS_APP_INFO(appInfo) || dstAppDesktopNameList == nullptr) {
        return;
    }

    const std::string appDesktopId = g_app_info_get_id(appInfo);
    QString localDesktopFile =
        QStandardPaths::locate(QStandardPaths::ApplicationsLocation, QString::fromStdString(appDesktopId));

    dstAppDesktopNameList->emplace_back(localDesktopFile.toStdString());
}

}

std::string AppChooser::getDefaultAppForUrl(const std::string &url)
{
    QUrl qUrl(QString::fromStdString(url));
    if (!qUrl.isValid()) {
        return std::string();
    }

    if (qUrl.isLocalFile()) {
        return getDefaultAppForLocalFile(qUrl);
    }

    return getDefaultAppForUrlScheme(qUrl.scheme().toStdString());
}

std::vector<std::string> AppChooser::getAvailableAppListForFile(const std::string &fileName)
{
    const std::string contentType = getMimeContentTypeFromFile(fileName);
    GList *appList = g_app_info_get_all_for_type(contentType.c_str());

    if (appList == nullptr) {
        return {};
    }

    std::vector<std::string> availableAppList;
    g_list_foreach(appList, eachAppInfoCallback, static_cast<gpointer>(&availableAppList));

    g_list_free_full (appList, g_object_unref);

    return availableAppList;
}

std::string AppChooser::getDefaultAppForLocalFile(const QUrl &url)
{
    auto contentType = getMimeContentTypeFromFile(url.url().toStdString());
    GAppInfo *appInfo = g_app_info_get_default_for_type(contentType.c_str(), false);

    if (appInfo == nullptr) {
        return std::string();
    }

    QString appId = g_app_info_get_id(appInfo);
    QString desktopFile = QStandardPaths::locate(QStandardPaths::ApplicationsLocation, appId);

    g_object_unref(appInfo);

    return desktopFile.toStdString();
}

std::string AppChooser::getDefaultAppForUrlScheme(const std::string &urlScheme)
{
    GAppInfo *appInfo = g_app_info_get_default_for_uri_scheme(urlScheme.c_str());
    QString appId = g_app_info_get_id(appInfo);
    QString desktopFile = QStandardPaths::locate(QStandardPaths::ApplicationsLocation, appId);

    g_object_unref(appInfo);

    return desktopFile.toStdString();
}

std::string AppChooser::getMimeContentTypeFromFile(const std::string &fileName)
{
    QMimeDatabase mimeDatabase;
    QMimeType mimeType = mimeDatabase.mimeTypeForFile(QString::fromStdString(fileName));
    return mimeType.name().toStdString();
}
