#ifndef APPINFOMANAGER_H
#define APPINFOMANAGER_H

#include <QTimer>
#include <map>
#include <vector>
#include <string>
#include "appinfo.h"
#include "desktopfilemanager.h"
#include "configmanager.h"

class AppInfoManager
{
public:
    using AppChangedCallback = std::function<void(const AppInfo &)>;

    explicit AppInfoManager(const std::shared_ptr<ConfigManager> &configManager);

    void handleWindowAdded(uint wid);
    void handleWindowRemoved(uint wid);
    void handleActiveWindowChanged(uint wid);
    void handleWindowMinimized(uint wid);
    void handleExceptionGroupSessionApp(const std::string &desktopFile, int pid);
    void handleMprisDbusAdded(int pid, const std::string &service);

    void createAppInstance(const std::string &desktopFile, int pid);
    void setAppGroupName(const std::string &appId, const std::string &groupName);

    void setAppStartedCallback(AppChangedCallback callback);
    void setAppStateChangedCallback(AppChangedCallback callback);
    void setAppChildPidsUpdatedCallback(AppChangedCallback callback);

    bool forceChangeAppStateByPid(int pid, sched_policy::AppState state);
    void forceChangCachedStateTo(sched_policy::AppState state);

    sched_policy::AppType getAppTypeByDesktopFile(const std::string &desktopFile);
    sched_policy::AppType getAppTypeByPid(int pid);

    std::string syncGetDesktopFileByPid(int pid);

    std::vector<AppInfo> getAllAppInfos();
    std::vector<AppInfo> getAppInfosWithState(sched_policy::AppState appState);

    AppInfo getLatestAppInfoByDesktopFile(
        const std::string &desktopFile, const std::vector<std::string> &args);
    AppInfo getLatestAppInfoByCmdline(const std::string &cmdline);
    AppInfo getAppInfo(const std::string &appId);
    std::string getDesktopFileByPid(int pid);

    void setBackgroundToCachedStateInterval(int interval);

private:
    using AppInfoIterator = std::vector<AppInfo>::iterator;

    void initActiveWindowChangedTimer();
    void doHandleActiveWindowChanged();
    void addWidToAppInfo(AppInfoIterator appInfoIt, uint wid);
    void updateAppCmdlineByWid(AppInfoIterator appInfoIt, uint wid);
    void createAppInstanceWithWid(uint wid);
    void appBecomeActiveWithWid(uint wid);
    void appBecomeInactiveWithWid(uint wid);

    void updateAppStateWithoutEmptyWindows(AppInfoIterator appInfoIt);
    void updateAppInfoWithEmptyWindows(AppInfoIterator appInfoIt);
    void setAppState(sched_policy::AppState appState, AppInfoIterator appInfoIt);
    unsigned long long setBackgroundStateTimer(const std::string &appId);
    void resetBackgroundStateTimers();
    std::vector<std::string> appIdsWithTimer();

    AppInfo generateAppInfoByWid(uint wid);
    AppInfo generateAppInfo(const std::string &desktopFile, int pid);

    using FindAppInfoPredicateCallback = std::function<bool(const AppInfo &)>;
    std::tuple<bool, AppInfoIterator> findAppInfo(const FindAppInfoPredicateCallback &predicateCallback);
    std::tuple<bool, AppInfoIterator> findAppInfoWithWid(uint wid);
    std::tuple<bool, AppInfoIterator> findAppInfoWithPid(int pid);
    std::tuple<bool, AppInfoIterator> findAppInfoWithAppId(const std::string &appId);

    void addStateChangedTimerId(const std::string &appId, unsigned long long timerId);
    void deleteStateChangeTimerId(const std::string &appId);

    void saveMprisServiceToCache(int pid, const std::string &service);
    std::string takeMprisServiceFromCache(int pid);

private:
    // <desktopfile id, appinfo list>
    std::map<std::string, std::vector<AppInfo>> m_appInfos;
    //<appId, timerId>
    std::map<std::string, unsigned long long> m_appStateChangeTimerIds;
    std::map<int, std::string> m_mprisServices;
    DesktopFileManager m_desktopFileManager;
    std::shared_ptr<ConfigManager> m_configManager;

    uint m_currentActiveWid;
    uint m_preActiveWid;
    QTimer m_activeWindowChangedTimer;

    int m_backgroundToCachedStateInterval;

    AppChangedCallback m_appStartedCallback;
    AppChangedCallback m_appStateChangedCallback;
    AppChangedCallback m_appChildPidsUpdatedCallback;
};

#endif // APPINFOMANAGER_H
