#include "batteryremainingtimeestimator.h"
#include <QDebug>

BatteryRemainingTimeEstimator::BatteryRemainingTimeEstimator(
    std::shared_ptr<SystemEnergy> systemEnergy,
    std::shared_ptr<ProcessGroupConsumptionMonitor> processGroupConsumptionMonitor,
    int estimateInterval,
    FrozenProcessGroupIdsGetter frozenProcessGroupIdsGetter)
    : m_systemEnergy(systemEnergy)
    , m_processGroupConsumptionMonitor(processGroupConsumptionMonitor)
    , m_estimateInterval(estimateInterval)
    , m_frozenProcessGroupIdsGetter(std::move(frozenProcessGroupIdsGetter))
{
    QObject::connect(&m_timer, &QTimer::timeout, [this]() {
        estimateRemainingTime();
    });
    m_timer.setInterval(m_estimateInterval * 1000);
}

void BatteryRemainingTimeEstimator::setIncreasedRemainTimeChangedCallback(
    RemainingTimeChangedCallback callback)
{
    m_increasedRemainingTimeChangedCallback = std::move(callback);
}

void BatteryRemainingTimeEstimator::startEstimate()
{
    qDebug() << "Battery remaning time estimator start.";
    if (!m_timer.isActive()) {
        m_timer.start();
        estimateRemainingTime();    
    }
}

void BatteryRemainingTimeEstimator::stopEstimate()
{
    qDebug() << "Battery remaning time estimator stop.";
    m_timer.stop();
    setIncreasedRemainingTime(0);
}

void BatteryRemainingTimeEstimator::estimateRemainingTime()
{
    std::vector<std::string> frozenGroupIds = m_frozenProcessGroupIdsGetter();
    if (frozenGroupIds.empty()) {
        setIncreasedRemainingTime(0);
        return;
    }

    double frozenGroupRate = calculateFrozenGroupsEnergyRate(frozenGroupIds);
    double estimatedEnergyRate = m_systemEnergy->dischargeRate() - frozenGroupRate;
    double currentEnergy = m_systemEnergy->energy();
    int batteryRemainingTime = currentEnergy / estimatedEnergyRate;
    int increasedTime = batteryRemainingTime > m_systemEnergy->timeToEmpty() ?
        batteryRemainingTime - m_systemEnergy->timeToEmpty() : 0;

    setIncreasedRemainingTime(increasedTime);
}

double BatteryRemainingTimeEstimator::calculateFrozenGroupsEnergyRate(
    const std::vector<std::string> &groupIds)
{
    double energyRate = 0.0;
    for (const auto &id : groupIds) {
        energyRate += m_processGroupConsumptionMonitor->cpuEnergyConsumptionRate(id);
    }

    return energyRate;
}

void BatteryRemainingTimeEstimator::setIncreasedRemainingTime(int time)
{
    if (m_increasedRemainingTimeChangedCallback) {
        m_increasedRemainingTimeChangedCallback(time);
    }
}
