#include "cgroupmanager.h"
#include <QCoreApplication>
#include <QFile>
#include <QDebug>
#include "misc.h"

#define LOG_WARNING(str) qWarning() << __FUNCTION__ << str << cgroup_strerror(cgroup_get_last_errno());

namespace {

const char *cleaner_path = "/usr/bin/kylin-process-manager-cleaner";
const char *freeze_controller = "freezer";
const char *cgroup_root_path = "/sys/fs/cgroup/";

void cgroupDeleter(cgroup *cg) {
    // Free internal cgroup structure. This function frees also all controllers
    // attached to the cgroup, including all parameters and their values.
    if (cg) { cgroup_free(&cg); }
};

}

bool CGroupManager::createProcessGroup(
    const QString &groupPath, const QStringList &controllers, const QList<int> &pids)
{
    auto cgroup = createCgroup(groupPath, controllers);
    if (cgroup == nullptr) {
        LOG_WARNING(QString("createCgroup %1 failed.").arg(groupPath));
        return false;
    }

    for (auto &pid : qAsConst(pids)) {
        int ret = cgroup_attach_task_pid(cgroup.get(), pid);
    }

    if (misc::version::cgroupVersion() == 2 &&
        groupPath.endsWith(".scope")) {
        m_cgroupv2ReleaseNotification.addWatchGroup(
            cgroup_root_path +  groupPath.toStdString(), 
            [this](std::string groupPath) {
                groupPath.erase(0, strlen(cgroup_root_path));
                removeProcessCGroup(QString::fromStdString(groupPath));
            });
    }
    
    return true;
}

bool CGroupManager::moveProcessToGroup(
    const QString &groupPath, const QStringList &controllers, const QList<int> &pids)
{
    return createProcessGroup(groupPath, controllers, pids);
}

bool CGroupManager::setProcessGroupResourceLimit(
    const QString &groupPath, const QString &controllerName, 
    const QString &attrFile, const QString &value)
{
    auto cgroup = initCgroupWithPath(groupPath);
    if (cgroup == nullptr)  {
        return false;
    }

    QByteArray controllerBa = controllerName.toLocal8Bit();
    // only v2. \todo supported v1
    const char *reallControllerName = 
        controllerName == freeze_controller ? "misc" : controllerBa.data();
        
    cgroup_controller *controller = cgroup_add_controller(cgroup.get(), reallControllerName);
    if (!controller) {
        LOG_WARNING(QString("cgroup_add_controller failed to path %1.").arg(groupPath));
        return false;
    }

    int ret = cgroup_set_value_string(
        controller, attrFile.toLocal8Bit().data(), value.toLocal8Bit().data());
    
    if (ret != 0) {
        LOG_WARNING(QString("cgroup_set_value_string %1 failed.").arg(groupPath));
        
        return false;
    }
    
    ret = cgroup_modify_cgroup(cgroup.get());

    if (ret != 0) {
        LOG_WARNING(QString("cgroup_modify_cgroup %1 failed: %2.").arg(groupPath, cgroup_strerror(ret)));
        qWarning() << "Controller name: " << reallControllerName 
                   <<  attrFile.toLocal8Bit().data() << value.toLocal8Bit().data();
        return false;
    }

    return true;
}

void CGroupManager::removeProcessCGroup(const QString &groupPath)
{
    qWarning() << "removeProcessCGroup" << groupPath;
    auto cgroup = initCgroupWithPath(groupPath);
    if (cgroup == nullptr)  {
        qWarning() << "removeProcessCGroup error: " << groupPath;
        return;
    }

    int ret = cgroup_get_cgroup(cgroup.get());
    if (ret != 0) {
        qDebug() << cgroup_strerror(ret);
        LOG_WARNING(QString("cgroup_get_cgroup %1 faild.").arg(groupPath));
        return;
    }

    ret = cgroup_delete_cgroup(cgroup.get(), 1);
    if (ret != 0) {
        LOG_WARNING(QString("cgroup_delete_cgroup_ext %1 faild.").arg(groupPath));
    }
}

CGroupManager::CgroupUniquePtr CGroupManager::initCgroupWithPath(const QString &groupPath)
{
    if (cgroup_init() != 0) {
        LOG_WARNING(QString("cgroup_init %1 failed.").arg(groupPath));
        return nullptr;
    }

    CgroupUniquePtr cgroup(cgroup_new_cgroup(groupPath.toLatin1().data()), cgroupDeleter);

    if (cgroup == nullptr) {
        LOG_WARNING(QString("cgroup_new_cgroup %1 failed.").arg(groupPath));
        return nullptr;
    }
    return cgroup;
}

CGroupManager::CgroupUniquePtr CGroupManager::createCgroup(
    const QString &cgroupName, const QStringList &controllers)
{
    auto cgroup = initCgroupWithPath(cgroupName);
    if (cgroup == nullptr)  {
        return nullptr;
    }

    if (!controllers.isEmpty()) {
        if (misc::version::cgroupVersion() == 1) {
            QString notifyOnReleaseValue =
            cgroupName.endsWith(".scope") && 
            cgroupName.contains("session.slice") ? "1" : "0";
            initControllers(cgroup.get(), controllers, notifyOnReleaseValue);
            setCGroupV1ReleaseAgent(controllers);
        } else {
            initControllers(cgroup.get(), controllers);
        }
    }

    if (cgroup_create_cgroup(cgroup.get(), 0) != 0) {
        LOG_WARNING(QString("cgroup_create_cgroup %1 failed.").arg(cgroupName));
        return nullptr;
    }

    return cgroup;
}

void CGroupManager::initControllers(
    cgroup *cgroup, const QStringList &controllers, const QString &notifyOnReleaseValue)
{
    for (auto &controllerName : qAsConst(controllers)) {
        if (controllerName == freeze_controller) {
            continue;
        }

        cgroup_controller *controller = 
            cgroup_add_controller(cgroup, controllerName.toLocal8Bit().data());
        if (!controller) {
            LOG_WARNING("cgroup_add_controller failed.");
            continue;
        }

        if (!notifyOnReleaseValue.isEmpty()) {
            int ret = cgroup_set_value_string(
                controller, "notify_on_release", 
                notifyOnReleaseValue.toLocal8Bit().data());
            if (ret != 0) {
                LOG_WARNING("set notify_on_release failed.");
            }
        } 
    }
}

void CGroupManager::setCGroupV1ReleaseAgent(const QStringList &controllers)
{
    auto cgroup = initCgroupWithPath("/");
    if (cgroup == nullptr)  {
        return;
    }

    for (auto &controlName : qAsConst(controllers)) {
        cgroup_controller *controller = 
            cgroup_add_controller(cgroup.get(), controlName.toLocal8Bit().data());
        if (!controller) {
            LOG_WARNING(QString("cgroup_add_controller %1 failed.").arg(controlName));
            continue;
        }
        int ret = cgroup_set_value_string(controller, "release_agent", cleaner_path);
        if (ret != 0) {
            LOG_WARNING(QString("set release_agent %1 failed.").arg(controlName));
        }
        if (cgroup_create_cgroup(cgroup.get(), 0) != 0) {
            LOG_WARNING(QString("cgroup_create_cgroup %1 failed.").arg(controlName));
        }   
    }
}

void CGroupManager::writePidToProcsFile(int pid, const QString &procsFile)
{
    QFile file(procsFile);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Unbuffered)) {
        qWarning() << "Write pid to cgroup.procs file failed, " << file.errorString();
        return;
    }
    file.write(QString::number(pid).toLocal8Bit());
}

QList<int> CGroupManager::getGroupPids(const QString &groupName) const
{
    if (cgroup_init() != 0) {
        fprintf(stderr, "Init failed\n");
        return QList<int>();
    }

    const char *controller = "pids";
    int size = 0;
    pid_t *pids = nullptr;
    QByteArray ba = groupName.toLocal8Bit();

    if (cgroup_get_procs(ba.data(), (char *)controller, &pids, &size) != 0) {
        qWarning() << "cgroup_get_procs error, " << cgroup_strerror(cgroup_get_last_errno());
        return QList<int>();
    }

    QList<int> result;
    for (int i = 0; i < size; ++i) {
        result.push_back(pids[i]);
    }

    delete[] pids;
    return result;
}
